# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import asyncio
import dataclasses
import os.path
import re
import typing
from collections.abc import Collection
from datetime import datetime
from urllib.parse import unquote

import httpx

from kitsunekko_tools.config import get_config
from kitsunekko_tools.consts import *
from kitsunekko_tools.ignore import IgnoreList


def file_exists(file_path: str) -> bool:
    return os.path.isfile(file_path) and os.stat(file_path).st_size > 0


@dataclasses.dataclass(frozen=True)
class AnimeSubtitleUrl:
    url: str
    title: str

    @property
    def file_name(self) -> str:
        return os.path.basename(unquote(self.url))


@dataclasses.dataclass(frozen=True)
class FetchResult:
    to_visit: set[str]
    to_download: set[AnimeSubtitleUrl]

    @classmethod
    def new(cls):
        return cls(to_visit=set(), to_download=set())

    def update(self, other: "FetchResult"):
        self.to_visit.update(other.to_visit)
        self.to_download.update(other.to_download)


@dataclasses.dataclass(frozen=True)
class DownloadResult:
    reason: str
    file_path: str

    def __repr__(self):
        return f"{self.reason}: {self.file_path}"


@dataclasses.dataclass(frozen=True)
class FetchState:
    to_visit: set[str]
    visited: set[str]

    @classmethod
    def new(cls, download_root: str) -> typing.Self:
        return cls(
            to_visit={
                download_root,
            },
            visited=set(),
        )

    def balance(self, prev_result: FetchResult) -> None:
        self.visited.update(self.to_visit)
        self.to_visit.clear()
        self.to_visit.update(prev_result.to_visit - self.visited)

    def has_unvisited(self) -> bool:
        return len(self.to_visit) > 0


@dataclasses.dataclass(frozen=True)
class DownloadError(Exception):
    url: str

    @property
    def errname(self) -> str:
        return type(self.__cause__).__name__


def get_anime_title(page_text: str) -> str:
    title = re.search(r"<title>([^<>]+)</title>", page_text, flags=re.IGNORECASE | re.MULTILINE).group(1)
    title = title.replace(" - Japanese subtitles - kitsunekko.net", "").replace("/", " ").replace("\\", " ")
    return title.strip()


def find_all_subtitle_dir_urls(html_text: str) -> list[str]:
    return re.findall(r'(?<="/)dirlist.php\?[^"\']+(?=")', html_text)


def find_all_subtitle_file_urls(html_text: str) -> list[str]:
    return re.findall(r'(?<=href=")subtitles/[^"\']+\.(?:zip|rar|ass|srt)(?=")', html_text)


async def crawl_page(client: httpx.AsyncClient, url: str) -> FetchResult:
    try:
        r = await client.get(url)
    except Exception as e:
        raise DownloadError(url) from e

    return FetchResult(
        to_visit={f"{KITSUNEKKO_DOMAIN_URL}/{anime_dir}" for anime_dir in find_all_subtitle_dir_urls(r.text)},
        to_download={
            AnimeSubtitleUrl(f"{KITSUNEKKO_DOMAIN_URL}/{subtitle}", get_anime_title(r.text))
            for subtitle in find_all_subtitle_file_urls(r.text)
        },
    )


async def find_subs_all(client: httpx.AsyncClient, to_visit: set[str]) -> FetchResult:
    result = FetchResult.new()
    for fut in asyncio.as_completed([crawl_page(client, page) for page in to_visit]):
        try:
            result.update(await fut)
        except DownloadError as e:
            print(f"got {e.errname} while trying to download {e.url}")
    return result


class Sync:
    def __init__(self):
        self._config = get_config().data
        self._ignore = IgnoreList(self._config)

    async def download_sub(self, client: httpx.AsyncClient, subtitle: AnimeSubtitleUrl) -> DownloadResult:
        if not os.path.isdir(dir_path := os.path.join(self._config.destination, subtitle.title)):
            os.mkdir(dir_path)
        if file_exists(file_path := os.path.join(dir_path, subtitle.file_name)):
            return DownloadResult("already exists", file_path)
        if self._ignore.is_matching(file_path):
            return DownloadResult("explicitly ignored", file_path)
        try:
            r = await client.get(subtitle.url)
        except Exception as e:
            raise DownloadError(subtitle.url) from e
        if r.status_code != httpx.codes.OK:
            return DownloadResult("download failed", file_path)
        with open(file_path, "wb") as f:
            f.write(r.content)
        return DownloadResult("saved", file_path)

    async def download_subs(self, client: httpx.AsyncClient, to_download: Collection[AnimeSubtitleUrl]) -> None:
        for fut in asyncio.as_completed([self.download_sub(client, subtitle) for subtitle in to_download]):
            try:
                print(await fut)
            except DownloadError as e:
                print(f"got {e.errname} while trying to download {e.url}")

    async def sync_all(self) -> None:
        async with httpx.AsyncClient(
            proxies=self._config.proxy,
            headers=self._config.headers,
            timeout=self._config.timeout,
        ) as client:
            state = FetchState.new(self._config.download_root)
            while state.has_unvisited():
                task: FetchResult = await find_subs_all(client, state.to_visit)
                print(f"visited {len(state.to_visit)} pages, found {len(task.to_download)} files.")
                await self.download_subs(client, task.to_download)
                state.balance(task)

        with open(os.path.join(self._config.destination, UPDATED_FILENAME), "w") as of:
            of.write(datetime.utcnow().strftime("%c"))



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import dataclasses
import fnmatch
import os.path
import pathlib
import sys
import typing

from kitsunekko_tools.common import KitsuException
from kitsunekko_tools.config import get_config, KitsuConfig
from kitsunekko_tools.consts import *


@dataclasses.dataclass
class IgnoreListException(KitsuException):
    what: str


class IgnoreList:
    _config: KitsuConfig
    _ignore_filepath: pathlib.Path

    def __init__(self, config: KitsuConfig | None = None):
        self._config = config or get_config().data
        self._ignore_filepath = pathlib.Path(self._config.destination) / IGNORE_FILENAME
        self._patterns = set()
        if not self._config.destination.is_dir():
            raise IgnoreListException(f"Destination directory does not exist: {self._config.destination}")
        try:
            with open(self._ignore_filepath, encoding="utf8") as f:
                self._patterns.update(filter(bool, map(str.strip, f.read().splitlines())))
        except FileNotFoundError:
            pass

    @property
    def ignore_filepath(self) -> pathlib.Path:
        return self._ignore_filepath

    def is_matching(self, file_path: str) -> bool:
        path_dest_stripped = os.path.relpath(file_path, self._config.destination)
        return any(fnmatch.fnmatch(path_dest_stripped, pattern) for pattern in self._patterns)

    def patterns(self) -> set[str]:
        """
        Return all known ignore patterns.
        """
        return self._patterns

    def add(self, pattern: str):
        """
        Add a new ignore pattern to the list.
        """
        self._patterns.add(pattern)
        self._ignore_filepath.write_text("\n".join(self._patterns))

    def path(self) -> pathlib.Path:
        """
        Return path to ignore file.
        """
        return self._ignore_filepath



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import abc


class KitsuException(Exception, abc.ABC):
    @abc.abstractmethod
    def what(self):
        raise NotImplementedError()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import fire

from kitsunekko_tools.common import KitsuException
from kitsunekko_tools.config import ConfigFileNotFoundError, KitsuConfig, get_config
from kitsunekko_tools.download import Sync
from kitsunekko_tools.ignore import IgnoreList, IgnoreListException
from kitsunekko_tools.mega_upload import MegaError, mega_upload


class ConfigCli:
    """
    Manage config file.
    """

    @staticmethod
    def create() -> None:
        """
        Create an example config file in the default location, if it does not exist.
        """
        try:
            data, location = get_config()
        except ConfigFileNotFoundError:
            config_file_path = KitsuConfig.default_location()
            config_file_path.parent.mkdir(exist_ok=True, parents=True)
            config_file_path.write_text(KitsuConfig().as_toml_str())
            print(f"Created config file: {config_file_path}")
        else:
            print(f"File already exists: {location}")

    @staticmethod
    def locate() -> None:
        """
        Print path to the config file, if it exists.
        """
        try:
            data, location = get_config()
        except ConfigFileNotFoundError as ex:
            print(ex.what)
        else:
            print(location)

    @staticmethod
    def show() -> None:
        """
        Show the content of the config file, if it exists.
        """
        try:
            data, location = get_config()
        except ConfigFileNotFoundError as ex:
            print(ex.what)
        else:
            print(data.as_toml_str())


class IgnoreCli:
    """
    Manage the list of ignore patterns.
    """

    def __init__(self, ignore_list: IgnoreList):
        self._ignore_list = ignore_list

    def locate(self) -> None:
        """
        Print path to the ignore file.
        """
        print(self._ignore_list.ignore_filepath)

    def show(self) -> None:
        """
        Print the list of ignore rules as Unix shell-style wildcards.
        """
        print("\n".join(self._ignore_list.patterns()))

    def add(self, pattern: str) -> None:
        """
        Add a new ignore rule.
        """
        if pattern:
            self._ignore_list.add(str(pattern))
            print(f"File written: {self._ignore_list.path()}")
        else:
            print("Nothing to add.")


class Application:
    """
    A set of scripts for creating a local kitsunekko mirror.
    """

    @staticmethod
    def config() -> ConfigCli:
        """
        Manage config file.
        """
        return ConfigCli()

    @staticmethod
    def destination() -> None:
        """
        Print path to destination directory.
        """
        try:
            data, location = get_config()
        except ConfigFileNotFoundError as ex:
            print(ex.what)
        else:
            print(data.destination)

    @staticmethod
    def ignore():
        """
        Manage the list of ignore patterns.
        """
        try:
            return IgnoreCli(IgnoreList())
        except IgnoreListException as ex:
            print(ex.what)

    @staticmethod
    async def sync():
        """
        Download everything from Kitsunekko to a local folder.
        """
        try:
            s = Sync()
        except KitsuException as ex:
            print(ex.what)
        else:
            await s.sync_all()

    @staticmethod
    def upload():
        """
        Upload the local folder to mega.nz.
        The ~/.megarc file must exist.
        """
        try:
            mega_upload(get_config().data)
        except KitsuException as ex:
            print(ex.what)


def main() -> None:
    fire.Fire(Application)


if __name__ == "__main__":
    main()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import dataclasses
import fnmatch
import functools
import io
import os.path
import pathlib
import sys
import tomllib
import typing

from kitsunekko_tools.common import KitsuException
from kitsunekko_tools.consts import *

DEFAULT_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; rv:108.0) Gecko/20100101 Firefox/108.0",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Charset": "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
    "Accept-Encoding": "none",
    "Accept-Language": "en-US,en;q=0.8",
    "Connection": "keep-alive",
}


@functools.cache
def get_xdg_config_dir() -> pathlib.Path:
    return pathlib.Path(os.environ.get("XDG_CONFIG_HOME", pathlib.Path.home() / ".config"))


@functools.cache
def config_locations():
    return (
        get_xdg_config_dir() / PROG_NAME / SETTINGS_FILE_NAME,
        pathlib.Path.home() / SETTINGS_FILE_NAME,
        pathlib.Path("/etc/") / PROG_NAME / SETTINGS_FILE_NAME,
    )


class ConfigFileNotFoundError(KitsuException, FileNotFoundError):
    @property
    def what(self) -> str:
        with io.StringIO() as si:
            si.write("Couldn't find config file. ")
            si.write("Create the file in one of the following locations:\n")
            si.write("\n".join(f"・ {location}" for location in config_locations()))
            return si.getvalue()


@dataclasses.dataclass
class KitsuConfig:
    destination: pathlib.Path = "/mnt/archive/japanese/kitsunekko-mirror"
    proxy: str = "socks5://127.0.0.1:9050"
    download_root: str = "https://kitsunekko.net/dirlist.php?dir=subtitles/japanese/"
    timeout: int = 120
    headers: dict[str, str] = dataclasses.field(default_factory=lambda: DEFAULT_HEADERS.copy())

    def __post_init__(self):
        self.destination = pathlib.Path(self.destination)

    def as_toml_str(self) -> str:
        with io.StringIO() as si:
            for key, value in dataclasses.asdict(self).items():
                match value:
                    case str() | pathlib.Path():
                        si.write(f'{key} = "{value}"\n')
                    case int():
                        si.write(f"{key} = {value}\n")
                    case dict():
                        si.write(f"[{key}]\n")
                        si.write("\n".join(f'{k} = "{v}"' for k, v in value.items()))
                    case _:
                        raise RuntimeError(f"Unknown value type {type(value)}")
            return si.getvalue()

    @staticmethod
    def default_location() -> pathlib.Path:
        return config_locations()[0]


class ReadConfigResult(typing.NamedTuple):
    data: KitsuConfig
    file_path: pathlib.Path


@functools.cache
def get_config() -> ReadConfigResult:
    for config_file_path in config_locations():
        if os.path.isfile(config_file_path):
            with open(config_file_path, "rb") as f:
                return ReadConfigResult(KitsuConfig(**tomllib.load(f)), config_file_path)
    raise ConfigFileNotFoundError()



# Copyright: Ajatt-Tools and contributors; https://github.com/Ajatt-Tools
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import pathlib as _pathlib

PROG_NAME = "kitsunekko-tools"
SETTINGS_FILE_NAME = "kitsunekko-tools.toml"
KITSUNEKKO_DOMAIN_URL = "https://kitsunekko.net"
REPO = _pathlib.Path(__file__).parent.absolute()
IGNORE_FILENAME = ".kitsuignore"
UPDATED_FILENAME = ".updated"

__all__ = [name for name in globals() if name.isupper()]
